/*
 * IGMP Snooping Layer: IGMP Snooping module runs at layer 2. IGMP
 * Snooping layer uses the multicast information in the IGMP messages
 * exchanged between the participating hosts and multicast routers to
 * update the multicast forwarding database. This file contains the
 * common code routines of IGS module.
 *
 * Copyright 2007, Broadcom Corporation
 * All Rights Reserved.                
 *                                     
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Broadcom Corporation;   
 * the contents of this file may not be disclosed to third parties, copied
 * or duplicated in any form, in whole or in part, without the prior      
 * written permission of Broadcom Corporation.                            
 *
 * $Id$
 */

#include <bcmdefs.h>
#include <bcmendian.h>
#include <proto/bcmip.h>
#include <osl.h>
#include <bcmnvram.h>
#include <clist.h>
#if defined(linux)
#include <osl_linux.h>
#else /* defined(osl_xx) */
#error "Unsupported osl"
#endif /* defined(osl_xx) */
#include "igs_cfg.h"
#include "emfc_export.h"
#include "igs_export.h"
#include "igsc_export.h"
#include "igsc.h"
#include "igsc_sdb.h"

typedef struct mc_grp_spl
{
	uint32	addr;
	uint32	mask;
} mc_grp_spl_t;

static mc_grp_spl_t mc_addr_rsvd[] =
{
	{ 0xe0000001, 0xffffffff },		/* All hosts */
	{ 0xe0000002, 0xffffffff },		/* All routers */
	{ 0xe0000004, 0xffffffff },		/* DVMRP routers */
	{ 0xe0000005, 0xffffffff },		/* OSPF1 routers */
	{ 0xe0000006, 0xffffffff },		/* OSPF2 routers */
	{ 0xe0000009, 0xffffffff },		/* RIP v2 routers */
	{ 0xe000000d, 0xffffffff },		/* PIMd routers */
	{ 0xe0000010, 0xffffffff },		/* IGMP v3 routers */
	{ 0xe00000fb, 0xffffffff },		/* Reserved */
	{ 0xe000ff87, 0xffffffff },		/* Reserved */
	{ 0xeffffffa, 0xffffffff },		/* UPnP */
};

static bool
igsc_is_reserved(uint32 addr)
{
	int32 i;

	for (i = 0; i < sizeof(mc_addr_rsvd)/sizeof(mc_grp_spl_t); i++)
	{
		if ((addr & mc_addr_rsvd[i].mask) ==
		    (mc_addr_rsvd[i].addr & mc_addr_rsvd[i].mask))
		{
			return (TRUE);
		}
	}

	return (FALSE);
}

/*
 * Description: This function is called by EMFL when an IGMP frame
 *              is received. Based on the IGMP packet type it either
 *              adds, deletes or refreshes the MFDB entry.
 *
 * Input:       s       - Snooper global instance data cookie
 *              ifp     - Pointer to interface the frame arrived on.
 *              iph     - Points to start of IP header in the packet.
 *              igmph   - Points to start of IGMP header in the packet.
 *              from_rp - TRUE when received from router port.
 *
 * Return:      EMF_FLOOD  - Flood the frame.
 *              EMF_SENDUP - Deliver the packet to IP layer.
 */
static int32
igsc_input(emfc_snooper_t *s, void *ifp, uint8 *iph, uint8 *igmph, bool from_rp)
{
	int32 ret = FAILURE;
	igsc_info_t *igsc_info;
	uint32 mgrp_ip, mh_ip, dest_ip;

	ASSERT(s != NULL);

	igsc_info = IGSC_INFO(s);

	IGSC_STATS_INCR(igsc_info, igmp_packets);

	if (from_rp)
	{
		IGS_DEBUG("Ignoring IGMP frame from router port\n");
		IGSC_STATS_INCR(igsc_info, igmp_not_handled);
		return (EMF_FLOOD);
	}

	mgrp_ip = ntoh32(*((uint32 *)(igmph + IGMPV2_GRP_ADDR_OFFSET)));
	mh_ip = ntoh32(*((uint32 *)(iph + IPV4_SRC_IP_OFFSET)));

	switch (igmph[IGMPV2_TYPE_OFFSET])
	{
		case IGMPV2_HOST_NEW_MEMBERSHIP_REPORT:
			/* Ignore frames received with reserved addresses */
			dest_ip = ntoh32(*((uint32 *)(iph + IPV4_DEST_IP_OFFSET)));
			if (igsc_is_reserved(dest_ip))
			{
				IGSC_STATS_INCR(igsc_info, igmp_not_handled);
				IGS_DEBUG("Reserved multicast address %x frame\n",
				          dest_ip);
				return (EMF_FLOOD);
			}

			/* When a host membership report is received add/refresh the
			 * entry for the host.
			 */
			IGSC_STATS_INCR(igsc_info, igmp_v2reports);
			ret = igsc_sdb_member_add(igsc_info, ifp, mgrp_ip, mh_ip);

			return ((ret == SUCCESS) ? EMF_SENDUP : EMF_FLOOD);

		case IGMPV2_LEAVE_GROUP_MESSAGE:
			/* Remove group entry for the host when a leave message is 
			 * received.
			 */
			IGSC_STATS_INCR(igsc_info, igmp_leaves);
			ret = igsc_sdb_member_del(igsc_info, ifp, mgrp_ip, mh_ip);

			if (ret != SUCCESS)
				IGS_WARN("Deleting unknown member with grp %x host %x if %p",
				         mgrp_ip, mh_ip, ifp);
			break;

		case IGMPV2_HOST_MEMBERSHIP_REPORT:
			IGSC_STATS_INCR(igsc_info, igmp_reports);
			break;

		case IGMPV2_HOST_MEMBERSHIP_QUERY:
			IGSC_STATS_INCR(igsc_info, igmp_queries);
			return (EMF_SENDUP);

		default:
			IGS_WARN("IGMP type %d not handled\n", igmph[IGMPV2_TYPE_OFFSET]);
			IGSC_STATS_INCR(igsc_info, igmp_not_handled);
			break;
	}

	return (EMF_FLOOD);
}

/*
 * Internet checksum routine
 */
static uint16
inet_cksum(uint8 *buf, int32 len)
{
	uint32 sum = 0;
	uint16 val;

	while (len > 1)
	{
		val = *buf++ << 8;
		val |= *buf++;
		sum += val;
		len -= 2;
	}

	if (len > 0)
	{
		sum += (*buf) << 8;
	}

	while (sum >> 16)
	{
		sum = (sum & 0xffff) + (sum >> 16);
	}

	return (hton16(~sum));
}

/*
 * Description: This function adds the IP header, IGMP header and data.
 */
static void
igsc_igmp_pkt_encap(uint8 *ip, uint32 src_ip, uint32 dest_ip,
                    uint8 igmp_type, uint8 max_rsp, uint32 grp_addr)
{
	uint8 *igmp;

	/* Prepare the IP header */
	ip[IP_VER_OFFSET] = ((IP_VER_4 << 4) | (IPV4_MIN_HLEN >> 2));
	ip[IPV4_TOS_OFFSET] = 0xc0;
	*((uint16 *)(ip + IPV4_LEN_OFFSET)) = hton16(IPV4_MIN_HLEN + IGMP_HLEN);
	*((uint16 *)(ip + IPV4_ID_OFFSET)) = 0;
	*((uint16 *)(ip + IPV4_FRAG_OFFSET)) = 0;
	ip[IPV4_TTL_OFFSET] = 1;
	ip[IPV4_PROT_OFFSET] = IP_PROT_IGMP;
	*((uint16 *)(ip + IPV4_CHKSUM_OFFSET)) = 0;
	*((uint32 *)(ip + IPV4_SRC_IP_OFFSET)) = src_ip;
	*((uint32 *)(ip + IPV4_DEST_IP_OFFSET)) = dest_ip;
	*((uint16 *)(ip + IPV4_CHKSUM_OFFSET)) = inet_cksum(ip, IPV4_MIN_HLEN);

	/* Fill the IGMP header fields */
	igmp = ip + IPV4_MIN_HLEN;
	igmp[IGMPV2_TYPE_OFFSET] = igmp_type;
	igmp[IGMPV2_MAXRSP_TIME_OFFSET] = max_rsp;
	*((uint16 *)(igmp + IGMPV2_CHECKSUM_OFFSET)) = 0;
	*((uint32 *)(igmp + IGMPV2_GRP_ADDR_OFFSET)) = grp_addr;
	*((uint16 *)(igmp + IGMPV2_CHECKSUM_OFFSET)) = inet_cksum(ip,
	                                               IGMP_HLEN + IPV4_MIN_HLEN);

	return;
}

/*
 * Description: This functions generates an IGMP Query. It is used to
 *              quickly determine the group members on the attached bridge
 *              ports.
 *
 * Return:      SUCCESS or FAILURE
 */
static int32
igsc_igmp_query_send(igsc_info_t *igsc_info)
{
	uint32 ip[64];
	int8 lan_ipaddr[16];
	uint32 lan_ip;
	uint32 lan_ip_byte1, lan_ip_byte2, lan_ip_byte3, lan_ip_byte4;

	/* Read the local IP address from NVRAM */
	strncpy(lan_ipaddr, nvram_safe_get("lan_ipaddr"), 16);
	lan_ipaddr[15] = 0;
	sscanf(lan_ipaddr, "%d.%d.%d.%d",
	       &lan_ip_byte1, &lan_ip_byte2, &lan_ip_byte3, &lan_ip_byte4);
	lan_ip = (lan_ip_byte1 | lan_ip_byte2 << 8 | lan_ip_byte3 << 16 |
	          lan_ip_byte4 << 24);

	/* Build the IGMP Query packet */
	igsc_igmp_pkt_encap((uint8 *)ip, lan_ip, hton32(IPV4_MCADDR_ALLHOSTS),
	                    IGMPV2_HOST_MEMBERSHIP_QUERY, IGMPV2_MAXRSP_TIME, 0);

	IGS_DEBUG("Forwarding IGMP Query on to bridge ports\n");

	return (igs_broadcast(igsc_info->igs_info, (uint8 *)ip,
	                      IGMP_HLEN + IPV4_MIN_HLEN,
	                      IPV4_MCADDR_ALLHOSTS));
}

/*
 * Description: This function is called by the EMFL when forwarding
 *              is enabled.
 *
 * Input:       snooper - Snooper gloabl instance data.
 *
 * Return:      SUCCESS or FAILURE
 */
static int32
igsc_emf_enabled(emfc_snooper_t *snooper)
{
	IGS_INFO("EMF enable callback called\n");

	/* Send query when in router mode only */
	if (nvram_match("router_disable", "1"))
		return (SUCCESS);

	/* Send an IGMP Query packet to quickly learn the members present
	 * in the network.
	 */
	if (igsc_igmp_query_send(IGSC_INFO(snooper)) != SUCCESS)
	{
		IGS_ERROR("IGMP Query send failed\n");
		return (FAILURE);
	}

	return (SUCCESS);
}

/*
 * Description: This function is called by the EMFL when forwarding
 *              is disabled.
 *
 * Input:       snooper - Snooper gloabl instance data.
 *
 * Return:      SUCCESS or FAILURE
 */
static int32
igsc_emf_disabled(emfc_snooper_t *snooper)
{
	IGS_INFO("EMF disable callback\n");
	igsc_sdb_clear(IGSC_INFO(snooper));
	return (SUCCESS);
}

/*
 * IGSL Packet Counters/Statistics Function
 */
int32
igsc_stats_get(igsc_info_t *igsc_info, igs_stats_t *stats, uint32 size)
{
	if (igsc_info == NULL)
	{
		IGS_ERROR("Invalid IGSC handle passed\n");
		return (FAILURE);
	}

	if (stats == NULL)
	{
		IGS_ERROR("Invalid buffer input\n");
		return (FAILURE);
	}

	if (size < sizeof(igs_stats_t))
	{
		IGS_ERROR("Insufficient buffer size %d\n", size);
		return (FAILURE);
	}

	*stats = igsc_info->stats;

	return (SUCCESS);
}

void
igsc_cfg_request_process(igsc_info_t *igsc_info, igs_cfg_request_t *cfg)
{
	IGS_DEBUG("IGS command identifier: %d\n", cfg->command_id);

	switch (cfg->command_id)
	{
		case IGSCFG_CMD_IGS_STATS:
			if (igsc_stats_get(igsc_info, (igs_stats_t *)cfg->arg,
			                   cfg->size) != SUCCESS)
			{
				cfg->status = IGSCFG_STATUS_FAILURE;
				cfg->size = sprintf(cfg->arg, "IGS stats get failed\n");
				break;
			}
			cfg->status = IGSCFG_STATUS_SUCCESS;
			break;

		case IGSCFG_CMD_IGSDB_LIST:
			if (igsc_sdb_list(igsc_info, (igs_cfg_sdb_list_t *)cfg->arg,
			                  cfg->size) != SUCCESS)
			{
				cfg->status = IGSCFG_STATUS_FAILURE;
				cfg->size = sprintf(cfg->arg, "IGSDB listing failed\n");
				break;
			}
			cfg->status = IGSCFG_STATUS_SUCCESS;
			break;

		default:
			cfg->status = IGSCFG_STATUS_CMD_UNKNOWN;
			cfg->size = sprintf(cfg->arg, "Unknown command id %d\n",
			                    cfg->command_id);
			break;
	}
}

/*
 * Description: This function is called from the OS specific module
 *              init routine to initialize the IGSL. This function
 *              primarily initializes the IGSL global data and IGSDB.
 *
 * Input:       inst_id    - IGS instance identifier.
 *              igs_info  - IGSL OS Specific global data handle
 *              osh       - OS abstraction layer handle
 *
 * Return:      igsc_info - IGSL Common code global data handle
 */
void *
igsc_init(int8 *inst_id, void *igs_info, osl_t *osh)
{
	igsc_info_t *igsc_info;

	if (inst_id == NULL)
	{
		IGS_ERROR("Instance identifier NULL\n");
		return (NULL);
	}

	IGS_DEBUG("Initializing IGMP Snooping Layer\n");

	/* Allocate memory */
	igsc_info = MALLOC(osh, sizeof(igsc_info_t));
	if (igsc_info == NULL)
	{
		IGS_ERROR("Failed to allocated memory size %d for MFL\n",
		          sizeof(igsc_info_t));
		return (NULL);
	}

	IGS_DEBUG("Allocated memory for IGSC info\n");

	/* Initialize the IGS global data */
	bzero(igsc_info, sizeof(igsc_info_t));
	igsc_info->osh = osh;
	igsc_info->igs_info = igs_info;
	igsc_info->grp_mem_intv = IGMPV2_GRP_MEM_INTV;

	/* Initialize the IGSDB */
	igsc_sdb_init(igsc_info);

	igsc_info->sdb_lock = OSL_LOCK_CREATE("SDB Lock");

	if (igsc_info->sdb_lock == NULL)
	{
		igsc_sdb_clear(igsc_info);
		MFREE(osh, igsc_info, sizeof(igsc_info_t));
		return (NULL);
	}

	/* Register IGMP v2 Snooper with EMFL */
	igsc_info->snooper.input_fn = igsc_input;
	igsc_info->snooper.emf_enable_fn = igsc_emf_enabled;
	igsc_info->snooper.emf_disable_fn = igsc_emf_disabled;

	if (emfc_igmp_snooper_register(inst_id, &igsc_info->emf_handle,
	                               &igsc_info->snooper) != SUCCESS)
	{
		IGS_ERROR("IGMP Snooper couldn't register with EMF\n");
		igsc_sdb_clear(igsc_info);
		OSL_LOCK_DESTROY(igsc_info->sdb_lock);
		MFREE(osh, igsc_info, sizeof(igsc_info_t));
		return (NULL);
	}

	IGS_DEBUG("Initialized IGSDB\n");

	return (igsc_info);
}

/*
 * Description: This function is called from OS specific module cleanup
 *              routine. This routine primarily stops the group interval
 *              timers and cleans up the IGSDB entries.
 */
void
igsc_exit(igsc_info_t *igsc_info)
{
	emfc_igmp_snooper_unregister(igsc_info->emf_handle);
	igsc_sdb_clear(igsc_info);
	OSL_LOCK_DESTROY(igsc_info->sdb_lock);
	MFREE(igsc_info->osh, igsc_info, sizeof(igsc_info_t));

	IGS_DEBUG("Cleaned up IGSL, exiting common code\n");

	return;
}
